/******************************************************************************
*  FFT Benchmark 1
******************************************************************************/

#define FFTW_FLAG FFTW_ESTIMATE
#define DELAY1 1
#define DELAY2 100
#define FACTOR 1

#include <algorithm>
#include <cstring>
#include <cmath>
#include <cstdlib>
#include "otfft/otfft.h"
#include "otfft/msleep.h"
#include "cpp_fftw3.h"
#include "ooura1.h"
#include "simple_fft.h"
//#include "simple_fft2.h"

using namespace std;
using OTFFT::complex_t;
using OTFFT::simd_malloc;
using OTFFT::simd_free;
using CppFFTW3::FORWARD;
using CppFFTW3::INVERSE;

#include "laptime.h"
#include "useriomanip.h"

void initialize(const int N, complex_t* x)
{
    for (int p = 0; p < N; p++) {
        const double t = double(p)/N;
        x[p].Re = 10 * cos(3*2*M_PI*t*t);
        x[p].Im = 10 * sin(3*2*M_PI*t*t);
    }
}

void usage()
{
    cerr << "usage: fftbench1" << endl;
    cerr << "       fftbench1 index" << endl;
    cerr << "       fftbench1 from_index to_index" << endl;
    cerr << "index: FFT size is 2^(index)" << endl;
    exit(0);
}

int main(int argc, char *argv[]) try
{
    const int n_min  = argc >= 2 ? atoi(argv[1]) : 1;
    const int n_max  = argc >= 3 ? atoi(argv[2]) : (argc == 2 ? n_min : 22);
    const int N_max  = 1 << n_max;
    const int nN_max = (max)(22*(1<<22), n_max*N_max);
    if (argc >= 2 && strncmp(argv[1], "-h", 2) == 0) usage();
    if (argc >= 2 && strncmp(argv[1], "--help", 6) == 0) usage();
    if (n_min < 1 || 24 < n_min) throw "index must be 1..24";
    if (n_max < 1 || 24 < n_max) throw "index must be 1..24";
    if (n_min > n_max) throw "index must be from_index <= to_index";

#ifndef DO_SINGLE_THREAD
    fftw_init_threads();
#endif
    cout << "------+-----------+-----------------+-----------------+-----------------+---\n";
    cout << "length|FFTW3[usec]|   OOURA   [usec]| SimpleFFT [usec]|   OTFFT   [usec]|err\n";
    cout << "------+-----------+-----------------+-----------------+-----------------+---\n";
    complex_t* x1 = (complex_t*) simd_malloc(N_max*sizeof(complex_t));
    complex_t* x2 = (complex_t*) simd_malloc(N_max*sizeof(complex_t));
    double sum1 = 0, sum2 = 0, sum3 = 0, sum4 = 0;
    for (int n = n_min; n <= n_max; n++) {
        const int N = 1 << n;
        const int LOOPS = nN_max/(n*N);
        const int TRIES = (min)(16, n*FACTOR);
        //const int TRIES = (min)(64, n*FACTOR);
        double lap, lap1;

        cout << "2^(" << fd(2, n) << ")|" << flush;

        //// FFTW3 ////
        initialize(N, x1);
        //if (n >= 20) fftw_set_timelimit(30);
        CppFFTW3::Plan<FORWARD,1,1> fwd_plan(N, x1);
        CppFFTW3::Plan<INVERSE,0,0> inv_plan(N, x1);
        lap = laptime1(LOOPS, TRIES, fwd_plan, inv_plan);
        sum1 += lap*n/n_max;
        cout << ff(11,2, lap/LOOPS) << "|" << flush;
        msleep(DELAY2);
        lap1 = lap;

        //// OOURA FFT ////
        initialize(N, x1);
        lap = laptime2(LOOPS, TRIES, OOURA::FFT(N), x1);
        sum2 += lap*n/n_max;
        cout << ff(11,2, lap/LOOPS) << "(" << ff(3,0, 100*lap/lap1) << "%)|" << flush;
        msleep(DELAY2);

        // erase OOURA FFT's influence
        //CppFFTW3::Plan<FORWARD,1,1> eraser(N, x1); eraser();

        //// Simple FFT ////
        initialize(N, x1);
        //OTFFT_MISC::speedup_magic();
        lap = laptime2(LOOPS, TRIES, SimpleFFT::FFT(N), x1);
        sum3 += lap*n/n_max;
        cout << ff(11,2, lap/LOOPS) << "(" << ff(3,0, 100*lap/lap1) << "%)|" << flush;
        msleep(DELAY2);

        //// OTFFT ////
        initialize(N, x1);
        //OTFFT_MISC::speedup_magic();
        lap = laptime2(LOOPS, TRIES, OTFFT::FFT(N), x1);
        sum4 += lap*n/n_max;
        cout << ff(11,2, lap/LOOPS) << "(" << ff(3,0, 100*lap/lap1) << "%)|" << flush;
        msleep(DELAY2);

        double err = 0;
        initialize(N, x1);
        initialize(N, x2);
        fwd_plan();
        //SimpleFFT::FFT(N).fwd(x1);
        OTFFT::FFT(N).fwd(x2);
        for (int k = 0; k < N; k++) {
            const complex_t d = x2[k] - x1[k];
            err += Re(d*conj(d));
        }
        initialize(N, x1);
        initialize(N, x2);
        inv_plan();
        //SimpleFFT::FFT(N).inv(x1);
        OTFFT::FFT(N).inv(x2);
        for (int k = 0; k < N; k++) {
            const complex_t d = x2[k] - x1[k];
            err += Re(d*conj(d));
        }
        if (err == 0)
            cout << " -" << endl;
        else if (err > 0)
            cout << ff(3,0, log10(err)) << endl;
        else
            cout << "NG" << endl;
        msleep(DELAY2);
    }
    simd_free(x2);
    simd_free(x1);
    cout << "------+-----------+-----------------+-----------------+-----------------+---\n";
    cout << " cost |";
    cout << ff(11,2, sum1/(n_max-n_min+1)) << "|" << flush;
    cout << ff(11,2, sum2/(n_max-n_min+1)) << "(" << ff(3,0, 100*sum2/sum1) << "%)|" << flush;
    cout << ff(11,2, sum3/(n_max-n_min+1)) << "(" << ff(3,0, 100*sum3/sum1) << "%)|" << flush;
    cout << ff(11,2, sum4/(n_max-n_min+1)) << "(" << ff(3,0, 100*sum4/sum1) << "%)|" << endl;
    cout << "------+-----------+-----------------+-----------------+-----------------+---\n";
#ifndef DO_SINGLE_THREAD
    fftw_cleanup_threads();
#endif

    return 0;
}
catch (const char *message) { cerr << message << endl; exit(1); }
catch (...) { cerr << "\n""*** exception! ***" << endl; exit(1); }
