/******************************************************************************
*  FFTW Benchmark
******************************************************************************/

#define FFTW_FLAG FFTW_MEASURE
#define DELAY1 1
#define DELAY2 100
#define FACTOR 1

#include <algorithm>
#include <cstring>
#include <cmath>
#include <cstdlib>
#include "otfft/otfft.h"
#include "otfft/msleep.h"
#include "cpp_fftw3.h"

using namespace std;
using OTFFT::complex_t;
using OTFFT::simd_malloc;
using OTFFT::simd_free;
using CppFFTW3::FORWARD;
using CppFFTW3::INVERSE;

#include "laptime.h"
#include "useriomanip.h"

void initialize(const int N, complex_t* x)
{
    for (int p = 0; p < N; p++) {
        const double t = double(p)/N;
        x[p].Re = 10 * cos(3*2*M_PI*t*t);
        x[p].Im = 10 * sin(3*2*M_PI*t*t);
    }
}

void usage()
{
    cerr << endl;
    cerr << "usage: fftwbench" << endl;
    cerr << "       fftwbench index" << endl;
    cerr << "       fftwbench from_index to_index" << endl;
    cerr << "       fftwbench from_index to_index import_wisdom" << endl;
    cerr << "       fftwbench from_index to_index import_wisdom export_wisdom" << endl;
    cerr << "index: FFT size is 2^(index)" << endl;
    cerr << "import_wisdom, export_wisdom: FFTW wisdom file" << endl;
    cerr << endl;
    exit(0);
}

int main(int argc, char *argv[]) try
{
    const int n_min = argc >= 2 ? atoi(argv[1]) : 1;
    const int n_max = argc >= 3 ? atoi(argv[2]) : (argc == 2 ? n_min : 22);
    const char* import_wisdom = argc >= 4 ? argv[3] : "wisdom";
    const char* export_wisdom = argc >= 5 ? argv[4] : "wisdom";
    const int N_max = 1 << n_max;
    const int nN_max = (max)(22*(1<<22), n_max*N_max);
    bool save_wisdom = false;
    if (argc >= 2 && strncmp(argv[1], "-h", 2) == 0) usage();
    if (argc >= 2 && strncmp(argv[1], "--help", 6) == 0) usage();
    if (argc == 5) save_wisdom = true;
    if (n_min < 1 || 24 < n_min) throw "index must be 1..24";
    if (n_max < 1 || 24 < n_max) throw "index must be 1..24";
    if (n_min > n_max) throw "index must be from_index <= to_index";

#ifndef DO_SINGLE_THREAD
    fftw_init_threads();
#endif
    if (fftw_import_wisdom_from_filename(import_wisdom) != 0)
        //cerr << "FFTW wisdom: import wisdom from '" << import_wisdom << "'." << endl;
        (void) 0;
    else
        save_wisdom = true;
    complex_t* x1 = (complex_t*) simd_malloc(N_max*sizeof(complex_t));
    //complex_t* x2 = (complex_t*) simd_malloc(N_max*sizeof(complex_t));
    double sum = 0;
    for (int n = n_min; n <= n_max; n++) {
        const int N = 1 << n;
        const int LOOPS = nN_max/(n*N);
        //const int TRIES = (min)(16, n*FACTOR);
        //const int TRIES = (min)(64, n*FACTOR);
        const int TRIES = n >= 14 ? 16 : (min)(20, n*FACTOR + 1);

        //complex_t* x1 = (complex_t*) simd_malloc(N*sizeof(complex_t));
        //complex_t* x2 = (complex_t*) simd_malloc(N*sizeof(complex_t));

        cout << "  " << fd(2, n) << " " << flush;

        //// FFTW3 ////
        //fftw_set_timelimit(180);
        initialize(N, x1);
        CppFFTW3::Plan<FORWARD,1,1> fwd_plan(N, x1);
        CppFFTW3::Plan<INVERSE,0,0> inv_plan(N, x1);
        double lap = laptime1(LOOPS, TRIES, fwd_plan, inv_plan);
        sum += lap*n/n_max;
        cout << ff(15,6, lap/LOOPS) << endl << flush;
        msleep(DELAY2);

        //simd_free(x2);
        //simd_free(x1);
    }
    //simd_free(x2);
    simd_free(x1);
    cout << "cost " << ff(15,6, sum/(n_max-n_min+1)) << endl;
    if (save_wisdom) {
        if (fftw_export_wisdom_to_filename(export_wisdom) == 0) {
            cerr << "FFTW wisdom: export error" << endl;
        }
    }
#ifndef DO_SINGLE_THREAD
    fftw_cleanup_threads();
#endif

    return 0;
}
catch (const char *message) { cerr << message << endl; exit(1); }
catch (...) { cerr << "\n""*** exception! ***" << endl; exit(1); }
