/******************************************************************************
*  OTFFT Benchmark
******************************************************************************/

#define DELAY1 1
#define DELAY2 100
#define FACTOR 2

#include <algorithm>
#include <cstring>
#include <cmath>
#include <cstdlib>
#include "otfft/otfft.h"
#include "otfft/otfft_misc.h"
#include "otfft/msleep.h"
#include "simple_fft.h"

using namespace std;
using OTFFT::complex_t;
using OTFFT::simd_malloc;
using OTFFT::simd_free;

#include "laptime.h"
#include "useriomanip.h"

void initialize(const int N, complex_t* x)
{
    for (int p = 0; p < N; p++) {
        const double t = double(p)/N;
        x[p].Re = 10 * cos(3*2*M_PI*t*t);
        x[p].Im = 10 * sin(3*2*M_PI*t*t);
    }
}

void usage()
{
    cerr << endl;
    cerr << "usage: otfftbench" << endl;
    cerr << "       otfftbench index" << endl;
    cerr << "       otfftbench from_index to_index" << endl;
    cerr << "       otfftbench from_index to_index -e" << endl;
    cerr << "index: FFT size is 2^(index)" << endl;
    cerr << "-e   : with error check" << endl;
    cerr << endl;
    exit(0);
}

int main(int argc, char *argv[]) try
{
    const int n_min = argc >= 2 ? atoi(argv[1]) : 1;
    const int n_max = argc >= 3 ? atoi(argv[2]) : (argc == 2 ? n_min : 22);
    const bool error_check = argc >= 4 ? strncmp(argv[3], "-e", 2) == 0 : false;
    const int N_max = 1 << n_max;
    const int nN_max = (max)(22*(1<<22), n_max*N_max);
    if (argc >= 2 && strncmp(argv[1], "-h", 2) == 0) usage();
    if (argc >= 2 && strncmp(argv[1], "--help", 6) == 0) usage();
    if (n_min < 1 || 24 < n_min) throw "index must be 1..24";
    if (n_max < 1 || 24 < n_max) throw "index must be 1..24";
    if (n_min > n_max) throw "index must be from_index <= to_index";

    complex_t* x1 = (complex_t*) simd_malloc(N_max*sizeof(complex_t));
    complex_t* x2 = (complex_t*) simd_malloc(N_max*sizeof(complex_t));
    double sum = 0;
    for (int n = n_min; n <= n_max; n++) {
        const int N = 1 << n;
        const int LOOPS = nN_max/(n*N);
        //const int TRIES = (min)(16, n*FACTOR);
        //const int TRIES = (min)(64, n*FACTOR);
        const int TRIES = n >= 14 ? 16 : (min)(20, n*FACTOR + 1);

        //complex_t* x1 = (complex_t*) simd_malloc(N*sizeof(complex_t));
        //complex_t* x2 = (complex_t*) simd_malloc(N*sizeof(complex_t));

        cout << "  " << fd(2, n) << " " << flush;

        initialize(N, x1);
        //OTFFT_MISC::speedup_magic();
        double lap = laptime2(LOOPS, TRIES, OTFFT::FFT(N), x1);
        sum += lap*n/n_max;
        cout << ff(15,6, lap/LOOPS) << flush;
        msleep(DELAY2);

        if (error_check) {
            double err = 0;
            initialize(N, x1);
            initialize(N, x2);
            SimpleFFT::FFT(N).fwd(x1);
            OTFFT::FFT(N).fwd(x2);
            for (int k = 0; k < N; k++) {
                const complex_t d = x2[k] - x1[k];
                err += Re(d*conj(d));
            }
            initialize(N, x1);
            initialize(N, x2);
            SimpleFFT::FFT(N).inv(x1);
            OTFFT::FFT(N).inv(x2);
            for (int k = 0; k < N; k++) {
                const complex_t d = x2[k] - x1[k];
                err += Re(d*conj(d));
            }
            if (err == 0)
                cout << " -";
            else if (err > 0)
                cout << " " << ff(3,0, log10(err));
            else
                cout << " NG";
            msleep(DELAY2);
        }

        cout << endl;

        //simd_free(x2);
        //simd_free(x1);
    }
    simd_free(x2);
    simd_free(x1);
    cout << "cost " << ff(15,6, sum/(n_max-n_min+1)) << endl;

    return 0;
}
catch (const char *message) { cerr << message << endl; exit(1); }
catch (...) { cerr << "\n""*** exception! ***" << endl; exit(1); }
